"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const commander_1 = require("commander");
const figures_1 = __importDefault(require("figures"));
const oas_1 = __importDefault(require("oas"));
const ora_1 = __importDefault(require("ora"));
const codegen_1 = __importDefault(require("../codegen"));
const fetcher_1 = __importDefault(require("../fetcher"));
const prompt_1 = __importDefault(require("../lib/prompt"));
const logger_1 = __importDefault(require("../logger"));
const storage_1 = __importDefault(require("../storage"));
// @todo log logs to `.api/.logs` and have `.logs` ignored
const cmd = new commander_1.Command();
cmd
    .name('install')
    .description('install an API SDK into your codebase')
    .argument('<uri>', 'an API to install')
    .option('-i, --identifier <identifier>', 'API identifier (eg. `@api/petstore`)')
    .addOption(new commander_1.Option('-l, --lang <language>', 'SDK language').choices([
    'js',
    'js-cjs',
    'js-esm',
    'ts',
]))
    .addOption(new commander_1.Option('-y, --yes', 'Automatically answer "yes" to any prompts printed'))
    .action(async (uri, options) => {
    let language;
    if (options.lang) {
        language = options.lang;
    }
    else {
        ({ value: language } = await (0, prompt_1.default)({
            type: 'select',
            name: 'value',
            message: 'What language would you like to generate an SDK for?',
            choices: [
                { title: 'TypeScript', value: 'ts' },
                { title: 'JavaScript', value: 'js' },
            ],
            initial: 1,
        }));
    }
    // Because our TS generation outputs raw TS source files we don't need to worry about CJS/ESM.
    if (language === 'js') {
        ({ value: language } = await (0, prompt_1.default)({
            type: 'select',
            name: 'value',
            message: 'How are your project imports and exports structured?',
            choices: [
                { title: 'CommonJS', description: 'require/exports', value: 'cjs' },
                { title: 'ECMAScript Modules', description: 'import/export', value: 'esm' },
            ],
            initial: 0,
            format: sel => (sel === 'cjs' ? 'js-cjs' : 'js-esm'),
        }));
    }
    // @todo let them know that we're going to be creating a `.api/ directory
    // @todo detect if they have a gitigore and .npmignore and if .api woudl be ignored by that
    // @todo don't support swagger files without upconverting them
    if (storage_1.default.isInLockFile({ source: uri })) {
        // @todo
        // logger(`It looks like you already have this API installed. Would you like to update it?`);
    }
    let identifier;
    if (options.identifier) {
        // `Storage.isIdentifierValid` will throw an exception if an identifier is invalid.
        if (storage_1.default.isIdentifierValid(options.identifier)) {
            identifier = options.identifier;
        }
    }
    else if (fetcher_1.default.isAPIRegistryUUID(uri)) {
        identifier = fetcher_1.default.getProjectPrefixFromRegistryUUID(uri);
    }
    else {
        ({ value: identifier } = await (0, prompt_1.default)({
            type: 'text',
            name: 'value',
            message: 'What would you like to identify this API as? This will be how you import the SDK. (e.g. entering `petstore` would result in `@api/petstore`)',
            validate: value => {
                if (!value) {
                    return false;
                }
                try {
                    return storage_1.default.isIdentifierValid(value, true);
                }
                catch (err) {
                    return err.message;
                }
            },
        }));
    }
    if (!identifier) {
        (0, logger_1.default)('You must tell us what you would like to identify this API as in order to install it.', true);
        process.exit(1);
    }
    let spinner = (0, ora_1.default)('Fetching your API').start();
    const storage = new storage_1.default(uri, identifier);
    const oas = await storage
        .load()
        .then(res => {
        spinner.succeed(spinner.text);
        return res;
    })
        .then(oas_1.default.init)
        .catch(err => {
        // @todo cleanup installed files
        spinner.fail(spinner.text);
        (0, logger_1.default)(err.message, true);
        process.exit(1);
    });
    // @todo look for a prettier config and if we find one ask them if we should use it
    spinner = (0, ora_1.default)('Generating your SDK').start();
    const generator = (0, codegen_1.default)(language, oas, './openapi.json', identifier);
    const sdkSource = await generator
        .generator()
        .then(res => {
        spinner.succeed(spinner.text);
        return res;
    })
        .catch(err => {
        // @todo cleanup installed files
        spinner.fail(spinner.text);
        (0, logger_1.default)(err.message, true);
        process.exit(1);
    });
    spinner = (0, ora_1.default)('Saving your SDK into your codebase').start();
    await storage
        .saveSourceFiles(sdkSource)
        .then(() => {
        spinner.succeed(spinner.text);
    })
        .catch(err => {
        // @todo cleanup installed files
        spinner.fail(spinner.text);
        (0, logger_1.default)(err.message, true);
        process.exit(1);
    });
    if (generator.hasRequiredPackages()) {
        (0, logger_1.default)(`${figures_1.default.warning} This generator requires some packages to be installed alongside it:`);
        Object.entries(generator.requiredPackages).forEach(([pkg, pkgInfo]) => {
            (0, logger_1.default)(`  ${figures_1.default.pointerSmall} ${pkg}: ${pkgInfo.reason} ${pkgInfo.url}`);
        });
        if (!options.yes) {
            await (0, prompt_1.default)({
                type: 'confirm',
                name: 'value',
                message: 'OK to proceed with package installation?',
                initial: true,
            }).then(({ value }) => {
                if (!value) {
                    // @todo cleanup installed files
                    (0, logger_1.default)('Installation cancelled.', true);
                    process.exit(1);
                }
            });
        }
        spinner = (0, ora_1.default)('Installing required packages').start();
        try {
            await generator.installer(storage);
            spinner.succeed(spinner.text);
        }
        catch (err) {
            // @todo cleanup installed files
            spinner.fail(spinner.text);
            (0, logger_1.default)(err.message, true);
            process.exit(1);
        }
    }
    (0, logger_1.default)('🚀 All done!');
})
    .addHelpText('after', `
Examples:
  $ api install @developers/v2.0#nysezql0wwo236
  $ api install https://raw.githubusercontent.com/readmeio/oas-examples/main/3.0/json/petstore-simple.json
  $ api install ./petstore.json`);
exports.default = cmd;
